@@ -162,6 +162,7 @@ let byte_offset_to_char_offset s byte_offset =
   done;
   byte_offset - !extra_bytes
 
+(* get byte length of a Unicode character *)
 let ulen uni_ch =
   if uni_ch < 0x80 then 1
   else if uni_ch < 0x800 then 2
@@ -484,6 +485,7 @@ let io_read_all chan =
 (** Run an external command asynchronously *)
 
 let run_command display finally cmd =
+  Minilib.log ("Start async command: " ^ cmd);
   let cin = Unix.open_process_in cmd in
   let fd = Unix.descr_of_in_channel cin in
   let () = Unix.set_nonblock fd in
