@@ -99,12 +99,12 @@ let reorder_context env sigma sign ord =
       | top::ord' when mem_q top moved_hyps ->
           let ((d,h),mh) = find_q top moved_hyps in
           if occur_vars_in_decl env sigma h d then
-            user_err ~hdr:"reorder_context"
+            user_err
               (str "Cannot move declaration " ++ Id.print top ++ spc() ++
               str "before " ++
               pr_sequence Id.print
                 (Id.Set.elements (Id.Set.inter h
-                  (global_vars_set_of_decl env sigma d))));
+                  (global_vars_set_of_decl env sigma d))) ++ str ".");
           step ord' expected ctxt_head mh (d::ctxt_tail)
       | _ ->
           (match ctxt_head with
@@ -134,8 +134,8 @@ let check_decl_position env sigma sign d =
   let needed = global_vars_set_of_decl env sigma d in
   let deps = dependency_closure env sigma (named_context_of_val sign) needed in
   if Id.List.mem x deps then
-    user_err ~hdr:"Logic.check_decl_position"
-      (str "Cannot create self-referring hypothesis " ++ Id.print x);
+    user_err
+      (str "Cannot create self-referring hypothesis " ++ Id.print x ++ str ".");
   x::deps
 
 (* Auxiliary functions for primitive MOVE tactic
@@ -182,6 +182,18 @@ let split_sign env sigma hfrom l =
   in
   splitrec [] l
 
+(* ocaml/ocaml#10027 triggered if inline record *)
+type cannot_move_hyp = { from : Id.t; hto : Id.t move_location; hyp : Id.t }
+exception CannotMoveHyp of cannot_move_hyp
+
+let () = CErrors.register_handler (function
+    | CannotMoveHyp { from; hto; hyp } ->
+      Some Pp.(str "Cannot move " ++ Id.print from ++
+               pr_move_location Id.print hto ++
+               str ": it occurs in the type of " ++
+               Id.print hyp ++ str ".")
+    | _ -> None)
+
 let move_hyp env sigma toleft (left,declfrom,right) hto =
   let open EConstr in
   let push prefix sign = List.fold_right push_named_context_val prefix sign in
@@ -196,11 +208,7 @@ let move_hyp env sigma toleft (left,declfrom,right) hto =
           if occur_vars_in_decl env sigma midvars d then
             if not (move_location_eq hto (MoveAfter hyp)) then
               (first, d :: middle, Id.Set.add hyp midvars)
-            else
-              user_err ~hdr:"move_hyp" (str "Cannot move " ++ Id.print (NamedDecl.get_id declfrom) ++
-                pr_move_location Id.print hto ++
-                str ": it occurs in the type of "
-                ++ Id.print hyp ++ str ".")
+            else raise (CannotMoveHyp {from = NamedDecl.get_id declfrom; hto; hyp})
           else
             (d::first, middle, midvars)
         in
@@ -221,11 +229,7 @@ let move_hyp env sigma toleft (left,declfrom,right) hto =
               let vars = global_vars_set_of_decl env sigma d in
               let depvars = Id.Set.union vars depvars in
               (first, d::middle, depvars)
-            else
-              user_err ~hdr:"move_hyp" (str "Cannot move " ++ Id.print (NamedDecl.get_id declfrom) ++
-                pr_move_location Id.print hto ++
-                str ": it depends on "
-                ++ Id.print hyp ++ str ".")
+            else raise (CannotMoveHyp {from = NamedDecl.get_id declfrom; hto; hyp})
           else
             (d::first, middle, depvars)
         in
@@ -316,11 +320,11 @@ let goal_type_of ~check env sigma c =
     (sigma, EConstr.Unsafe.to_constr t)
   else (sigma, EConstr.Unsafe.to_constr (Retyping.get_type_of env sigma (EConstr.of_constr c)))
 
+let mk_goal = Goal.V82.mk_goal [@ocaml.warning "-3"]
+(* TODO: inline when the Goal.V82 module is removed *)
+
 let rec mk_refgoals ~check env sigma goalacc conclty trm =
   let hyps = Environ.named_context_val env in
-  let mk_goal hyps concl =
-    Goal.V82.mk_goal sigma hyps concl
-  in
     if (not check) && not (occur_meta sigma (EConstr.of_constr trm)) then
       let t'ty = Retyping.get_type_of env sigma (EConstr.of_constr trm) in
       let t'ty = EConstr.Unsafe.to_constr t'ty in
@@ -332,7 +336,7 @@ let rec mk_refgoals ~check env sigma goalacc conclty trm =
         let conclty = nf_betaiota env sigma conclty in
           if check && occur_meta sigma conclty then
             raise (RefinerError (env, sigma, MetaInType conclty));
-          let (gl,ev,sigma) = mk_goal hyps conclty in
+          let (gl,ev,sigma) = mk_goal sigma hyps conclty in
           let ev = EConstr.Unsafe.to_constr ev in
           let conclty = EConstr.Unsafe.to_constr conclty in
           gl::goalacc, conclty, sigma, ev
@@ -402,12 +406,10 @@ let rec mk_refgoals ~check env sigma goalacc conclty trm =
 
 and mk_hdgoals ~check env sigma goalacc trm =
   let hyps = Environ.named_context_val env in
-  let mk_goal hyps concl =
-    Goal.V82.mk_goal sigma hyps concl in
   match kind trm with
     | Cast (c,_, ty) when isMeta c ->
         let sigma = check_typability ~check env sigma ty in
-        let (gl,ev,sigma) = mk_goal hyps (nf_betaiota env sigma (EConstr.of_constr ty)) in
+        let (gl,ev,sigma) = mk_goal sigma hyps (nf_betaiota env sigma (EConstr.of_constr ty)) in
         let ev = EConstr.Unsafe.to_constr ev in
         gl::goalacc,ty,sigma,ev
 
@@ -505,7 +507,7 @@ and treat_case ~check env sigma ci lbrty lf acc' =
            form [Meta] *)
         let head = strip_outer_cast head in
         if isMeta head then begin
-          assert (args = Context.Rel.to_extended_vect mkRel 0 ctx);
+          assert (args = Context.Rel.instance mkRel 0 ctx);
           let (r,_,s,head'') = mk_refgoals ~check env sigma lacc ty head in
           let fi' = it_mkLambda_or_LetIn (mkApp (head'',args)) ctx in
           (r,s,fi'::bacc)
@@ -528,10 +530,10 @@ let convert_hyp ~check ~reorder env sigma d =
   | d' ->
     let c = Option.map EConstr.of_constr (NamedDecl.get_value d') in
     if check && not (is_conv env sigma (NamedDecl.get_type d) (EConstr.of_constr (NamedDecl.get_type d'))) then
-      user_err ~hdr:"Logic.convert_hyp"
+      user_err
         (str "Incorrect change of the type of " ++ Id.print id ++ str ".");
     if check && not (Option.equal (is_conv env sigma) b c) then
-      user_err ~hdr:"Logic.convert_hyp"
+      user_err
         (str "Incorrect change of the body of "++ Id.print id ++ str ".");
     let sign' = apply_to_hyp sign id (fun _ _ _ -> EConstr.Unsafe.to_named_decl d) in
     if reorder then reorder_val_context env sigma sign' (check_decl_position env sigma sign d)
@@ -552,7 +554,14 @@ let refiner ~check r =
   let (sgl,cl',sigma,oterm) = mk_refgoals ~check env sigma [] cl r in
   let map gl = Proofview.goal_with_state gl st in
   let sgl = List.rev_map map sgl in
-  let sigma = Goal.V82.partial_solution env sigma (Proofview.Goal.goal gl) (EConstr.of_constr oterm) in
+  let evk = Proofview.Goal.goal gl in
+  let c = EConstr.of_constr oterm in
+  (* Check that the goal itself does not appear in the refined term *)
+  let _ =
+    if not (Evarutil.occur_evar_upto sigma evk c) then ()
+    else Pretype_errors.error_occur_check env sigma evk c
+  in
+  let sigma = Evd.define evk c sigma in
   Proofview.Unsafe.tclEVARS sigma <*>
   Proofview.Unsafe.tclSETGOALS sgl
   end
