@@ -29,6 +29,8 @@ open Context.Rel.Declaration
 (** Machinery to custom the behavior of the reduction *)
 module ReductionBehaviour = Reductionops.ReductionBehaviour
 
+type volatile = { volatile : bool } [@@unboxed]
+
 (** Machinery about stack of unfolded constants *)
 module Cst_stack = struct
   open EConstr
@@ -38,34 +40,34 @@ module Cst_stack = struct
 - constant applied to params = term in head applied to args
 - there is at most one arguments with an empty list of args, it must be the first.
 - in args, the int represents the indice of the first arg to consider *)
-  type t = (constr * constr list * (int * constr array) list)  list
+  type t = (constr * constr list * (int * constr array) list * volatile)  list
 
   let empty = []
   let is_empty = CList.is_empty
 
   let drop_useless = function
-    | _ :: ((_,_,[])::_ as q) -> q
+    | _ :: ((_,_,[],_)::_ as q) -> q
     | l -> l
 
   let add_param h cst_l =
     let append2cst = function
-      | (c,params,[]) -> (c, h::params, [])
-      | (c,params,((i,t)::q)) when i = pred (Array.length t) ->
-        (c, params, q)
-      | (c,params,(i,t)::q) ->
-        (c, params, (succ i,t)::q)
+      | (c,params,[],vol) -> (c, h::params, [], vol)
+      | (c,params,((i,t)::q),vol) when i = pred (Array.length t) ->
+        (c, params, q, vol)
+      | (c,params,(i,t)::q, vol) ->
+        (c, params, (succ i,t)::q, vol)
     in
       drop_useless (List.map append2cst cst_l)
 
   let add_args cl =
-    List.map (fun (a,b,args) -> (a,b,(0,cl)::args))
+    List.map (fun (a,b,args,vol) -> (a,b,(0,cl)::args,vol))
 
-  let add_cst cst = function
-    | (_,_,[]) :: q as l -> l
-    | l -> (cst,[],[])::l
+  let add_cst ?(volatile=false) cst = function
+    | (_,_,[],_) :: q as l -> l
+    | l -> (cst,[],[],{volatile})::l
 
   let best_cst = function
-    | (cst,params,[])::_ -> Some(cst,params)
+    | (cst,params,[],_)::_ -> Some(cst,params)
     | _ -> None
 
   let reference sigma t = match best_cst t with
@@ -78,7 +80,7 @@ module Cst_stack = struct
     let reconstruct_head = List.fold_left
       (fun t (i,args) -> mkApp (t,Array.sub args i (Array.length args - i))) in
     List.fold_right
-      (fun (cst,params,args) t -> Termops.replace_term sigma
+      (fun (cst,params,args,_) t -> Termops.replace_term sigma
         (reconstruct_head d args)
         (applist (cst, List.rev params))
         t) cst_l c
@@ -87,10 +89,11 @@ module Cst_stack = struct
     let open Pp in
     let p_c c = Termops.Internal.print_constr_env env sigma c in
     prlist_with_sep pr_semicolon
-      (fun (c,params,args) ->
+      (fun (c,params,args,{volatile}) ->
         hov 1 (str"(" ++ p_c c ++ str ")" ++ spc () ++ pr_sequence p_c params ++ spc () ++ str "(args:" ++
                  pr_sequence (fun (i,el) -> prvect_with_sep spc p_c (Array.sub el i (Array.length el - i))) args ++
-                 str ")")) l
+               str ")" ++
+              (if volatile then str " (volatile)" else mt()))) l
 end
 
 
@@ -112,7 +115,13 @@ sig
   | Proj of Projection.t * Cst_stack.t
   | Fix of ('a, 'a) pfixpoint * 'a t * Cst_stack.t
   | Primitive of CPrimitives.t * (Constant.t * EInstance.t) * 'a t * CPrimitives.args_red * Cst_stack.t
-  | Cst of cst_member * int * int list * 'a t * Cst_stack.t
+  | Cst of { const : cst_member;
+             curr : int;
+             remains : int list;
+             params : 'a t;
+             volatile : bool;
+             cst_l : Cst_stack.t;
+           }
 
   and 'a t = 'a member list
 
@@ -166,7 +175,13 @@ struct
   | Proj of Projection.t * Cst_stack.t
   | Fix of ('a, 'a) pfixpoint * 'a t * Cst_stack.t
   | Primitive of CPrimitives.t * (Constant.t * EInstance.t) * 'a t * CPrimitives.args_red * Cst_stack.t
-  | Cst of cst_member * int * int list * 'a t * Cst_stack.t
+  | Cst of { const : cst_member;
+             curr : int;
+             remains : int list;
+             params : 'a t;
+             volatile : bool;
+             cst_l : Cst_stack.t;
+           }
 
   and 'a t = 'a member list
 
@@ -188,7 +203,7 @@ struct
     | Primitive (p,c,args,kargs,cst_l) ->
       str "ZPrimitive(" ++ str (CPrimitives.to_string p)
       ++ pr_comma () ++ pr pr_c args ++ str ")"
-    | Cst (mem,curr,remains,params,cst_l) ->
+    | Cst {const=mem;curr;remains;params;cst_l} ->
       str "ZCst(" ++ pr_cst_member pr_c mem ++ pr_comma () ++ int curr
       ++ pr_comma () ++
         prlist_with_sep pr_semicolon int remains ++
@@ -249,9 +264,9 @@ struct
         f_fix f1 f2
           && equal_rec (List.rev s1) (List.rev s2)
           && equal_rec s1' s2'
-      | Cst (c1,curr1,remains1,params1,_)::s1', Cst (c2,curr2,remains2,params2,_)::s2' ->
-        equal_cst_member c1 c2
-          && equal_rec (List.rev params1) (List.rev params2)
+      | Cst c1::s1', Cst c2::s2' ->
+        equal_cst_member c1.const c2.const
+          && equal_rec (List.rev c1.params) (List.rev c2.params)
           && equal_rec s1' s2'
       | ((App _|Case _|Proj _|Fix _|Cst _|Primitive _)::_|[]), _ -> false
     in equal_rec (List.rev sk1) (List.rev sk2)
@@ -289,7 +304,7 @@ struct
   let will_expose_iota args =
     List.exists
       (function (Fix (_,_,l) | Case (_,l) |
-                 Proj (_,l) | Cst (_,_,_,_,l)) when Cst_stack.is_empty l -> true | _ -> false)
+                 Proj (_,l) | Cst {cst_l=l}) when Cst_stack.is_empty l -> true | _ -> false)
       args
 
   let list_of_app_stack s =
@@ -325,12 +340,13 @@ struct
   (** This function breaks the abstraction of Cst_stack ! *)
   let best_state sigma (_,sk as s) l =
     let rec aux sk def = function
-      |(cst, params, []) -> (cst, append_app_list (List.rev params) sk)
-      |(cst, params, (i,t)::q) -> match decomp sk with
+      |(_,_,_,{volatile=true}) -> def
+      |(cst, params, [], _) -> (cst, append_app_list (List.rev params) sk)
+      |(cst, params, (i,t)::q, vol) -> match decomp sk with
         | Some (el,sk') when EConstr.eq_constr sigma el t.(i) ->
           if i = pred (Array.length t)
-          then aux sk' def (cst, params, q)
-          else aux sk' def (cst, params, (succ i,t)::q)
+          then aux sk' def (cst, params, q, vol)
+          else aux sk' def (cst, params, (succ i,t)::q, vol)
         | _ -> def
     in List.fold_left (aux sk) s l
 
@@ -357,10 +373,10 @@ struct
       zip (best_state sigma (mkFix fix, st @ (append_app [|f|] s)) cst_l)
   | f, (Fix (fix,st,_)::s) -> zip
     (mkFix fix, st @ (append_app [|f|] s))
-  | f, (Cst (cst,_,_,params,cst_l)::s) when refold ->
-    zip (best_state sigma (constr_of_cst_member cst (params @ (append_app [|f|] s))) cst_l)
-  | f, (Cst (cst,_,_,params,_)::s) ->
-    zip (constr_of_cst_member cst (params @ (append_app [|f|] s)))
+  | f, (Cst {const;params;cst_l}::s) when refold ->
+    zip (best_state sigma (constr_of_cst_member const (params @ (append_app [|f|] s))) cst_l)
+  | f, (Cst {const;params}::s) ->
+    zip (constr_of_cst_member const (params @ (append_app [|f|] s)))
   | f, (Proj (p,cst_l)::s) when refold ->
     zip (best_state sigma (mkProj (p,f),s) cst_l)
   | f, (Proj (p,_)::s) -> zip (mkProj (p,f),s)
@@ -402,14 +418,15 @@ let safe_meta_value sigma ev =
 
 (* Beta Reduction tools *)
 
-let apply_subst recfun env sigma cst_l t stack =
+let apply_subst env sigma cst_l t stack =
   let rec aux env cst_l t stack =
     match (Stack.decomp stack, EConstr.kind sigma t) with
     | Some (h,stacktl), Lambda (_,_,c) ->
        let cst_l' = Cst_stack.add_param h cst_l in
        aux (h::env) cst_l' c stacktl
-    | _ -> recfun sigma cst_l (substl env t, stack)
-  in aux env cst_l t stack
+    | _ -> (cst_l, (substl env t, stack))
+  in
+  aux env cst_l t stack
 
 (* Iota reduction tools *)
 
@@ -445,10 +462,10 @@ let magically_constant_of_fixbody env sigma reference bd = function
                     let get u = Option.get (Universe.level u) in
                     get u, get v
                 in
-                Univ.LMap.add l r acc)
-                csts Univ.LMap.empty
+                Univ.Level.Map.add l r acc)
+                csts Univ.Level.Map.empty
             in
-            let inst = Instance.subst_fn (fun u -> Univ.LMap.find u subst) u in
+            let inst = Instance.subst_fn (fun u -> Univ.Level.Map.find u subst) u in
             mkConstU (cst, EInstance.make inst)
           | None -> bd
         end
@@ -470,11 +487,9 @@ let contract_cofix env sigma ?reference (bodynum,(names,types,bodies as typedbod
 let reduce_and_refold_cofix recfun env sigma cst_l cofix sk =
   let raw_answer =
     contract_cofix env sigma ?reference:(Cst_stack.reference sigma cst_l) cofix in
-  apply_subst
-    (fun sigma x (t,sk') ->
-      let t' = Cst_stack.best_replace sigma (mkCoFix cofix) cst_l t in
-      recfun x (t',sk'))
-    [] sigma Cst_stack.empty raw_answer sk
+  let (x, (t, sk')) = apply_subst [] sigma Cst_stack.empty raw_answer sk in
+  let t' = Cst_stack.best_replace sigma (mkCoFix cofix) cst_l t in
+  recfun x (t', sk')
 
 (* contracts fix==FIX[nl;i](A1...Ak;[F1...Fk]{B1....Bk}) to produce
    Bi[Fj --> FIX[nl;j](A1...Ak;[F1...Fk]{B1...Bk})] *)
@@ -499,11 +514,9 @@ let contract_fix env sigma ?reference ((recindices,bodynum),(names,types,bodies
 let reduce_and_refold_fix recfun env sigma cst_l fix sk =
   let raw_answer =
     contract_fix env sigma ?reference:(Cst_stack.reference sigma cst_l) fix in
-  apply_subst
-    (fun sigma x (t,sk') ->
-      let t' = Cst_stack.best_replace sigma (mkFix fix) cst_l t in
-      recfun x (t',sk'))
-    [] sigma Cst_stack.empty raw_answer sk
+  let (x, (t, sk')) = apply_subst [] sigma Cst_stack.empty raw_answer sk in
+  let t' = Cst_stack.best_replace sigma (mkFix fix) cst_l t in
+  recfun x (t', sk')
 
 module CredNative = Reductionops.CredNative
 
@@ -542,11 +555,11 @@ let apply_branch env sigma (ind, i) args (ci, u, pms, iv, r, lf) =
       List.rev args
     else
       let ctx = expand_branch env sigma u pms (ind, i) br in
-      subst_of_rel_context_instance ctx args
+      subst_of_rel_context_instance_list ctx args
   in
   Vars.substl subst (snd br)
 
-let rec whd_state_gen ?csts flags env sigma =
+let whd_state_gen ?csts flags env sigma =
   let open Context.Named.Declaration in
   let open ReductionBehaviour in
   let rec whrec cst_l (x, stack) =
@@ -613,17 +626,21 @@ let rec whd_state_gen ?csts flags env sigma =
                   || is_case tm'
                   then fold ()
                   else whrec cst_l' (tm', sk' @ sk)
-              | UnfoldWhen { recargs } -> (* maybe unfolds *)
+              | UnfoldWhen { recargs; nargs } -> (* maybe unfolds *)
                 begin match recargs with
-                  |[] -> (* if nargs has been specified *)
-                    (* CAUTION : the constant is NEVER refold
-                                            (even when it hides a (co)fix) *)
+                  | [] -> (* if nargs has been specified *)
+                    (* CAUTION : the constant is NEVER refolded (even when it hides a (co)fix) *)
                     whrec cst_l (body, stack)
-                  |curr::remains -> match Stack.strip_n_app curr stack with
+                  | curr :: remains -> match Stack.strip_n_app curr stack with
                     | None -> fold ()
                     | Some (bef,arg,s') ->
-                      whrec Cst_stack.empty
-                        (arg,Stack.Cst(Stack.Cst_const (fst const, u'),curr,remains,bef,cst_l)::s')
+                      let cst_l = Stack.Cst
+                          { const = Stack.Cst_const (fst const, u');
+                            volatile = Option.has_some nargs;
+                            curr; remains; params=bef; cst_l;
+                          }
+                      in
+                      whrec Cst_stack.empty (arg,cst_l::s')
                 end
               end
         end
@@ -655,27 +672,45 @@ let rec whd_state_gen ?csts flags env sigma =
                    let idx = x - npars in
                    if idx < 0 then None else Some idx) recargs
                in
+               let volatile = match behavior with
+                 | UnfoldWhen {nargs} -> Option.has_some nargs
+                 | UnfoldWhenNoMatch _ | NeverUnfold -> false
+               in
                match recargs with
-               |[] -> (* if nargs has been specified *)
+               | [] -> (* if nargs has been specified *)
                  (* CAUTION : the constant is NEVER refold
                                   (even when it hides a (co)fix) *)
                  let stack' = (c, Stack.Proj (p, cst_l) :: stack) in
                  whrec Cst_stack.empty(* cst_l *) stack'
-               | curr::remains ->
+               | curr :: remains ->
                  if curr == 0 then (* Try to reduce the record argument *)
-                   whrec Cst_stack.empty
-                     (c, Stack.Cst(Stack.Cst_proj p,curr,remains,Stack.empty,cst_l)::stack)
+                   let cst_l = Stack.Cst
+                       { const=Stack.Cst_proj p;
+                         volatile; curr; remains;
+                         params=Stack.empty;
+                         cst_l;
+                       }
+                   in
+                   whrec Cst_stack.empty (c, cst_l::stack)
                  else
                    match Stack.strip_n_app curr stack with
                    | None -> fold ()
                    | Some (bef,arg,s') ->
-                     whrec Cst_stack.empty
-                       (arg,Stack.Cst(Stack.Cst_proj p,curr,remains,
-                                      Stack.append_app [|c|] bef,cst_l)::s')
+                     let cst_l = Stack.Cst
+                         { const=Stack.Cst_proj p;
+                           curr;
+                           remains;
+                           volatile;
+                           params=Stack.append_app [|c|] bef;
+                           cst_l;
+                         }
+                     in
+                     whrec Cst_stack.empty (arg,cst_l::s')
            end)
 
     | LetIn (_,b,_,c) when CClosure.RedFlags.red_set flags CClosure.RedFlags.fZETA ->
-      apply_subst (fun _ -> whrec) [b] sigma cst_l c stack
+      let (cst_l, p) = apply_subst [b] sigma cst_l c stack in
+      whrec cst_l p
     | Cast (c,_,_) -> whrec cst_l (c, stack)
     | App (f,cl)  ->
       whrec
@@ -684,23 +719,8 @@ let rec whd_state_gen ?csts flags env sigma =
     | Lambda (na,t,c) ->
       (match Stack.decomp stack with
       | Some _ when CClosure.RedFlags.red_set flags CClosure.RedFlags.fBETA ->
-        apply_subst (fun _ -> whrec) [] sigma cst_l x stack
-      | None when CClosure.RedFlags.red_set flags CClosure.RedFlags.fETA ->
-        let env' = push_rel (LocalAssum (na, t)) env in
-        let whrec' = whd_state_gen flags env' sigma in
-        (match EConstr.kind sigma (Stack.zip ~refold:true sigma (whrec' (c, Stack.empty))) with
-        | App (f,cl) ->
-          let napp = Array.length cl in
-          if napp > 0 then
-            let (x', l') = whrec' (Array.last cl, Stack.empty) in
-            match EConstr.kind sigma x', l' with
-            | Rel 1, [] ->
-              let lc = Array.sub cl 0 (napp-1) in
-              let u = if Int.equal napp 1 then f else mkApp (f,lc) in
-              if noccurn sigma 1 u then (pop u,Stack.empty),Cst_stack.empty else fold ()
-            | _ -> fold ()
-          else fold ()
-        | _ -> fold ())
+        let (cst_l, p) = apply_subst [] sigma cst_l x stack in
+        whrec cst_l p
       | _ -> fold ())
 
     | Case (ci,u,pms,p,iv,d,lf) ->
@@ -726,7 +746,7 @@ let rec whd_state_gen ?csts flags env sigma =
           let x' = Stack.zip sigma (x, args) in
           let out_sk = s' @ (Stack.append_app [|x'|] s'') in
           reduce_and_refold_fix whrec env sigma cst_l f out_sk
-        |args, (Stack.Cst (const,curr,remains,s',cst_l) :: s'') ->
+        |args, (Stack.Cst {const;curr;remains;volatile;params=s';cst_l} :: s'') ->
           let x' = Stack.zip sigma (x, args) in
           begin match remains with
           | [] ->
@@ -737,20 +757,27 @@ let rec whd_state_gen ?csts flags env sigma =
               | Some body ->
                 let const = (fst const, EInstance.make (snd const)) in
                 let body = EConstr.of_constr body in
-                whrec (Cst_stack.add_cst (mkConstU const) cst_l)
-                  (body, s' @ (Stack.append_app [|x'|] s'')))
+                let cst_l = Cst_stack.add_cst ~volatile (mkConstU const) cst_l in
+                whrec cst_l (body, s' @ (Stack.append_app [|x'|] s'')))
             | Stack.Cst_proj p ->
               let stack = s' @ (Stack.append_app [|x'|] s'') in
-                match Stack.strip_n_app 0 stack with
-                | None -> assert false
-                | Some (_,arg,s'') ->
-                  whrec Cst_stack.empty (arg, Stack.Proj (p,cst_l) :: s''))
+              match Stack.strip_n_app 0 stack with
+              | None -> assert false
+              | Some (_,arg,s'') ->
+                whrec Cst_stack.empty (arg, Stack.Proj (p,cst_l) :: s''))
           | next :: remains' -> match Stack.strip_n_app (next-curr-1) s'' with
             | None -> fold ()
             | Some (bef,arg,s''') ->
-              whrec Cst_stack.empty
-                (arg,
-                 Stack.Cst (const,next,remains',s' @ (Stack.append_app [|x'|] bef),cst_l) :: s''')
+              let cst_l = Stack.Cst
+                  { const;
+                    curr=next;
+                    volatile;
+                    remains=remains';
+                    params=s' @ (Stack.append_app [|x'|] bef);
+                    cst_l;
+                  }
+              in
+              whrec Cst_stack.empty (arg, cst_l :: s''')
           end
         |_, (Stack.App _)::_ -> assert false
         |_, _ -> fold ()
